/* CEC (HDMI Consumer Electronic Control) Library
 Copyright (c) 2009 Michael Dreher <michael(at)5dot1.de>  All right reserved.
 
 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.
 
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.
 
 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */


#include "local_defs.h"

volatile static unsigned long CECPrevSignalChange; // the time when the last level change has been detected
volatile static unsigned long CECPrevEventStart;
volatile static bool CECPrevSignalState;
volatile static byte CECStateMachine;
volatile static byte CECSubStateMachine;
volatile unsigned short CECBitCount;
volatile byte CECmsg[16];
volatile byte CECmsgLen;
volatile static bool CECEOM; // end of message flag
volatile byte CECOwnAddress = 14;

#if CECDEBUG == 1
volatile byte CECErrNo;
volatile unsigned long CECErrInfo;
#define CEC_ERROR(micros, errNo, info) { CECErrNo = errNo; CECStateMachine = CEC_STATE_ERROR; CECErrInfo = info; }
#else
#define CEC_ERROR(micros, errNo, info) { ResetStateMachine(micros); }
#endif

void CEC_init(void)
{
  cbi(CEC_IN_PORT, CEC_IN_BIT_NR); // disable internal pullup for input
  cbi(CEC_IN_DDR, CEC_IN_BIT_NR); // input  

  sbi(CEC_OUT_DDR, CEC_OUT_BIT_NR); // output  
  cbi(CEC_OUT_PORT, CEC_OUT_BIT_NR); // output: HIGH (idle) state

  delay(200); // wait until power has stabilized and all capacitors are fully charged

  // enable pin change IRQ for the DATA pin, any change triggers the IRQ
  cli();
  DATA_PCMSK |= _BV(CEC_IN_BIT_NR);
  PCICR |= _BV(DATA_PCIE);
  startIrqCapture();
  sei();
}


void ResetStateMachine(unsigned long curMicros)
{
  CECPrevSignalState = false;  
  CECPrevSignalChange = curMicros;
  CECPrevEventStart = CECPrevSignalChange;
  CECBitCount = 0;
  CECmsgLen = 0;

  CECStateMachine = CEC_STATE_IDLE;
  CECSubStateMachine = 0;
}

void startIrqCapture(void)
{
  unsigned char sreg = SREG;
  cli();
  ResetStateMachine(micros());
  SREG = sreg;
}

// returns the following state of the state machine
byte PushBitValue(bool b)
{
  if(++CECBitCount == 9)
  {
    // EOM bit received
    CECBitCount = 0;
    CECmsgLen++;
    CECEOM = b;
    return CEC_STATE_RECEIVE_SENDACK_WAITFORLOW;
  }
  else
  {
    CECmsg[CECmsgLen] = (CECmsg[CECmsgLen] << 1) | b;
    return CEC_STATE_RECEIVE_WAITFORLO;
  }
}

// XXX TODO: implement timer
void SetAckStartTimer(void)
{
  // Header Block or Data Block unicast to our logical address
  if((CECmsg[0] & 0x0f) == CECOwnAddress)
  {
    CEC_OUT_PORT |= _BV(CEC_OUT_BIT_NR);
  }

  // XXX TODO: ACK for broadcast messages is not implemented
}

// XXX TODO: implement timer
void ClearAckStopTimer(void)
{
  // Header Block or Data Block unicast to our logical address
  if((CECmsg[0] & 0x0f) == CECOwnAddress)
  {
    CEC_OUT_PORT &= ~_BV(CEC_OUT_BIT_NR);
  }
}

ISR(DATA_INT_VECT_NAME)
{
  bool curSignalState = (CEC_IN_PIN >> CEC_IN_BIT_NR) & 0x01;

  if(curSignalState != CECPrevSignalState) // check if the signal has changed
  {
    setLED(curSignalState);
    CECPrevSignalState = curSignalState;
    unsigned long curMicros = micros();
    unsigned long timeRun = curMicros - CECPrevSignalChange;

    if(CECStateMachine == CEC_STATE_IDLE)
    {
      // make sure the signal has been low for at a while to avoid starting in the middle of a transmission
      if(curSignalState
        && (timeRun > CEC_SILENTTIME)) // minimum silence period
      {
        // first low->high edge detected
        CECStateMachine = CEC_STATE_RECEIVE;
        CECSubStateMachine = CEC_STATE_RECEIVE_WAITFORSTARTBITTIMEA;
        CECPrevEventStart = curMicros;
      }
      else
      {
        //CECPrevSignalState = false; // ignore high->low edges
        CEC_ERROR(curMicros, 1, 0);
      }
    }
    else if(CECStateMachine == CEC_STATE_RECEIVE)
    {
      unsigned long timeEvent = curMicros - CECPrevEventStart;
      if(CECSubStateMachine == CEC_STATE_RECEIVE_WAITFORSTARTBITTIMEA)
      {
        //if(timeEvent >= 3400 && timeEvent <= 4000)
        if(timeEvent >= 3500 && timeEvent <= 3900)
          CECSubStateMachine = CEC_STATE_RECEIVE_WAITFORSTARTBITTIMEB;
        else
          CEC_ERROR(curMicros, 2, timeEvent);
      }
      else if(CECSubStateMachine == CEC_STATE_RECEIVE_WAITFORSTARTBITTIMEB)
      {
        //if(timeEvent >= 4200 && timeEvent <= 4900)
        if(timeEvent >= 4300 && timeEvent <= 4700)
        {
          CECSubStateMachine = CEC_STATE_RECEIVE_WAITFORHI;
          CECPrevEventStart = curMicros;
        }
        else
          CEC_ERROR(curMicros, 3, timeEvent);
      }
      else if(CECSubStateMachine == CEC_STATE_RECEIVE_WAITFORHI)
      {
        if(timeEvent >= 400 && timeEvent <= 850) // between T1 and T3 => logical 1
          CECSubStateMachine = PushBitValue(1);
        else if(timeEvent >= 1250 && timeEvent <= 1700) // between T4 and T6 => logical 0
            CECSubStateMachine = PushBitValue(0);
        else
          CEC_ERROR(curMicros, 4, timeEvent);
      }
      else if(CECSubStateMachine == CEC_STATE_RECEIVE_WAITFORLO)
      {
        if(timeEvent >= 2050 && timeEvent <= 2750) // between T7 and T8 => end of bit
        {
          CECSubStateMachine = CEC_STATE_RECEIVE_WAITFORHI;
          CECPrevEventStart = curMicros;
        }
        else
          CEC_ERROR(curMicros, 5, timeEvent);
      }
      else if(CECSubStateMachine == CEC_STATE_RECEIVE_SENDACK_WAITFORLOW)
      {
        if(timeEvent >= 2050 && timeEvent <= 2750) // between T7 and T8 => end of bit
        {
          CECSubStateMachine = CEC_STATE_RECEIVE_SENDACK_WAITFORTIME;

          // XXX TODO: what happens with the signal changes, do they trigger the ISR again after we exit it?
          SetAckStartTimer();


          {
            { // XXX TODO: the delay should be done using an interrupt instead of busy waiting
              CECPrevSignalState = false;  // make sure the next detected level change is a hi->low edge
              delayMicroseconds(1400);
              ClearAckStopTimer();
              if(CECEOM)
              {
                CECStateMachine = CEC_STATE_RECEIVED;
                CECSubStateMachine = 0;
              }
              else
              {
                CECSubStateMachine = CEC_STATE_RECEIVE_WAITFORLO;
              }
            }
          }         

          CECPrevEventStart = curMicros;
        }
        else
          CEC_ERROR(curMicros, 6, timeEvent);
      }
      /* XXX TODO: not needed at the moment
       else if(CECSubStateMachine == CEC_STATE_RECEIVE_SENDACK_WAITFORTIME)
       {
       ClearAckStopTimer();
       if(CECEOM)
       {
       CECStateMachine = CEC_STATE_RECEIVED;
       CECSubStateMachine = 0;
       }
       else
       {
       CECSubStateMachine = CEC_STATE_RECEIVE_WAITFORLO;
       }
       }
       */
    }

    CECPrevSignalChange = curMicros;
  }
}


unsigned char checkCaptureFlags(void)
{
  return CECStateMachine;
}










